/********************************************************************************

Copyright: Acclaim Systems Ltd 2004
Author:    rskinner
 
No claims are made concerning the suitability of this source code for any purpose. 
All source is provided 'as is' without express or implied warranty of any kind.
Acclaim Systems Ltd disclaims any/all implied warranties of merchantability, 
non-infringement or fitness for any particular purpose. In no event will Acclaim
Systems Ltd be liable for any special, indirect, incidental, consequential or 
any other damages resulting from loss of use, data or profits arising out of or 
in connection with the use or performance of this source code.

License:   Free to use and abuse for non-profit activities so long as the 
           attribution and disclaimer above are not altered. For-profit use 
           is prohibited without express written permission.
 
*******************************************************************************/

#ifndef H2D_STRUCT_H
#define H2D_STRUCT_H

#include "h2d.h"

#include <boost/spirit/actor/push_back_actor.hpp>
#include <boost/spirit/actor/assign_actor.hpp>
#include <boost/bind.hpp>
#include <boost/function.hpp>

#include <string>
#include <vector>

//////////////////////////////////////////////////////////////////////////////////////////
//
// Stuff to deal with struct definitions and 'typedef struct's
// Also strips out extraneous enum definitions - the rules for these follow
// the same for as for struct/union. They may have to be stripped out into 
// a separate header at some point.
//

//#define PRINT_DBG 1  // Print some debug info

namespace h2d {

  // Parse closure - limited to 3 members unless I start messing with the 
  // spirit limit parameters. Means the code isn't as efficient as it could
  // (bit more string copying) but the compile times are kept down.
  //
  struct StructClosure : public boost::spirit::closure<StructClosure,
                         std::string,
                         std::string,
                         std::string> {
    member1 retval_;
    member2 type_;
    member3 name_;
  };

  struct StructGrammar : public boost::spirit::grammar<StructGrammar> {

    template<typename T>
    struct definition {

      typedef typename T::iterator_t iterator_t;
      typedef boost::spirit::rule<T> rule_t;
      typedef boost::spirit::rule<T, StructClosure::context_t> crule_t;
      
      definition(StructGrammar const& self) {

        ows_ = *space_p;
        mws_ = +space_p;
        name_ = +(alnum_p | '_');
        type_ = str_p("struct") | str_p("union");
        types_ = (!ch_p('*') >> name_);

        // Formulate the content rule for struct's - mimic the EchoGrammar
        // Note that the 2 chset_p's are different! 
        text_ = print_p - (space_p | chset_p("!\"`'%^&*()-+={}[]:;~?<>,./\\|"));
        content_ = (chset_p("!\"`'%^&*()-+=[]:;~?<>,./\\|")) | (+text_) | (+space_p);

        body_ = confix_p((ch_p('{') >> *space_p)[boost::bind(&definition<T>::begin, boost::ref(body_), _1, _2)], 
                           +((nested_[boost::bind(&definition<T>::echo, boost::ref(body_), _1)]) 
                             | 
                             (reference_[boost::bind(&definition<T>::echo, boost::ref(body_), _1)])
                             |
                             (content_[boost::bind(&definition<T>::echo, boost::ref(body_), _1, _2)])),
                           ch_p('}')[boost::bind(&definition<T>::end, boost::ref(body_))]);

        // Handle structures defined within other aggregates
        nested_ = 
          type_[boost::bind(&definition<T>::type, boost::ref(nested_), _1, _2, "nested")] >>
          mws_[boost::bind(&definition<T>::echo, boost::ref(nested_), _1, _2)] >> 
          // possible we'll get a tag for a nested aggregate - bin it... (might have to reconsider later)
          (!name_ >> ows_) >> 
          body_[boost::bind(&definition<T>::echo, boost::ref(nested_), _1)] >> 
          ows_[boost::bind(&definition<T>::echo, boost::ref(nested_), _1, _2)] >> 
          name_[boost::bind(&definition<T>::name, boost::ref(nested_), _1, _2)] >>
          (*(anychar_p - ch_p(';') >> ch_p(';')))[boost::bind(&definition<T>::emitNested, boost::ref(nested_), _1, _2)];

        // Handle internal references to structs/unions - basically only
        // definitions for nested structs are allowed to have the struct
        // keyword inside a struct definition, so if we find 'struct/union' 
        // we can just gobble it up.
        reference_ = 
          ((type_ | "enum") >> ~chset_p("a-zA-Z0-9_"));

        // Sometimes a C header will forward declare a struct type and then
        // immediately typedef it. Without this rule we would end up with 
        // the orginal forward decl plus another one for the typedef which
        // would then have to be patched (as D doesn't allow multiple declarations)
        rule_0_ = 
          type_[boost::bind(&definition<T>::type, boost::ref(rule_0_), _1, _2, "rule0")] >> +space_p >>
          ((+(alnum_p | '_'))[boost::bind(&definition<T>::name, boost::ref(rule_0_), _1, _2)]) >> *space_p >>
          ';' >> *space_p >>
          // Now for detecting the typedef...
          str_p("alias") >> mws_ >>
          // type - assume that the type in the typedef will match the forward decl
          type_ >> mws_ >> 
          // name - must match the one from the forward decl
          f_str_p(boost::bind(&definition<T>::nameBegin, boost::ref(rule_0_)), 
                  boost::bind(&definition<T>::nameEnd, boost::ref(rule_0_))) >> mws_ >> 
          types_[boost::bind(&definition<T>::types, boost::ref(rule_0_), _1, _2, true)] >> ows_ >>
          (*(ch_p(',') >> ows_ >> types_[boost::bind(&definition<T>::types, boost::ref(rule_0_), _1, _2, false)] >> ows_)) >>
          ch_p(';')[boost::bind(&definition<T>::emitGlobal, boost::ref(rule_0_))]; 

        // typedef struct [name] [{ ... }] <typedef>[, <typedef>[...]];
        // 'typedef' will have been translated to 'alias'
        rule_1_ = 
          str_p("alias") >> mws_ >>
          (type_ | str_p("enum"))[boost::bind(&definition<T>::type, boost::ref(rule_1_), _1, _2, "rule1")] >> mws_ >> 
          // Optional name - if there isn't one we'll substitute/generate one
          (!(name_[boost::bind(&definition<T>::name, boost::ref(rule_1_), _1, _2)] >> ows_)) >> 
          (!(body_[boost::bind(&definition<T>::echo, boost::ref(rule_1_), _1)] >> ows_)) >> 
          types_[boost::bind(&definition<T>::types, boost::ref(rule_1_), _1, _2, true)] >> ows_ >>
          (*(ch_p(',') >> ows_ >> types_[boost::bind(&definition<T>::types, boost::ref(rule_1_), _1, _2, false)] >> ows_)) >>
          ch_p(';')[boost::bind(&definition<T>::emitGlobal, boost::ref(rule_1_))]; 

        // Need to process struct declarations that are already in the
        // D compatible style so that nested structs/unions are sorted
        // struct abc { struct {...} xyz; ... };
        rule_2_ = 
          type_[boost::bind(&definition<T>::type, boost::ref(rule_2_), _1, _2, "rule3")] >> mws_ >> 
          name_[boost::bind(&definition<T>::name, boost::ref(rule_2_), _1, _2)] >> ows_ >> 
          body_[boost::bind(&definition<T>::echo, boost::ref(rule_2_), _1)] >> ows_ >> 
          ch_p(';')[boost::bind(&definition<T>::emitGlobal, boost::ref(rule_2_))]; 
          
        // Remove any instance of the struct/union/enum keywords that is not
        // followed by a definition block '{...}' - with an optional tag.
        // 'struct <name>;' is also allowed
        rule_3_ = 
          (((type_ | "enum") >> +space_p >> 
            *(alnum_p | '_') >> *space_p >> chset_p("{;"))[boost::bind(&definition<T>::emit, _1, _2)]) 
          |
          ((type_ | "enum") >> +space_p);

        // The full rule set
        rule_ = rule_0_ | rule_1_ | rule_2_ | rule_3_;
      }
      
      
      rule_t const& start() const { return rule_; }
      
    private:
      rule_t  ows_;
      rule_t  mws_;
      rule_t  type_;
      rule_t  text_;
      rule_t  content_;
      rule_t  name_;
      rule_t  types_;

      // Rules which carry a closure
      crule_t body_;
      crule_t nested_;
      crule_t reference_;
      crule_t rule_0_;
      crule_t rule_1_;
      crule_t rule_2_;
      rule_t  rule_3_;

      rule_t  rule_;
      
      static void emit(iterator_t begin, iterator_t end) {
        std::cout << std::string(begin, end);
      }
      
      static void type(crule_t const& rule, iterator_t begin, iterator_t end, const char* x) {
#ifdef PRINT_DBG
        std::cout << ":" << x << ":" << std::string(begin, end) << ":E:";
#endif
        rule.type_() = std::string(begin, end);
      }
      static void name(crule_t const& rule, iterator_t begin, iterator_t end) {
#ifdef PRINT_DBG
        std::cout << ":NBES:" << std::string(begin, end) << ":NBEE:";
#endif
        rule.name_() = std::string(begin, end);
      }
      static void echo(crule_t const& rule, iterator_t begin, iterator_t end) {
#ifdef PRINT_DBG
        std::cout << ":EBEB:" << std::string(begin, end) << ":EBEE:";
#endif
        rule.retval_() += std::string(begin, end);
      }
      static void echo(crule_t const& rule, std::string const& str) {
        rule.retval_() += str;
      }
      static void types(crule_t const& rule, iterator_t begin, iterator_t end, bool first) {
        if (first) {
          rule.retval_() += ";\n";
          if (rule.name_().empty()) {
            // little bit of processing to generate a name for the struct
            // can't just use the first typedef because it may be a pointer definition
            std::string td(begin, end);
            std::string::size_type pos = td.find('*');
            if (pos == std::string::npos) {
              // This isn't pointer typedef - name the struct after it
              rule.name_() = td;
              // and return early before we alias it to itself
              return;
            }
            else {
              // This is a pointer typedef - generate a composite name
              rule.name_() = rule.type_() + td.substr(pos + 1);
            } 
          }
        }
        rule.retval_() += std::string("alias ");
        rule.retval_() += rule.name_();
        rule.retval_() += ' ';
        rule.retval_() += std::string(begin, end);
        rule.retval_() += ";\n";
      }
      static void begin(crule_t const& rule, iterator_t begin, iterator_t end) {
#ifdef PRINT_DBG
        std::cout << ":BB:" << std::string(begin, end) << ":BE:";
#endif
        rule.retval_() += std::string(begin, end);
      }
      static void end(crule_t const& rule) {
#ifdef PRINT_DBG
        std::cout << ":EB:}:EE:";
#endif
        rule.retval_() += '}';
      }
      static void emitGlobal(crule_t const& rule) {
#ifdef PRINT_DBG
        std::cout << ":EGSB:";
#endif
        std::cout << 
          rule.type_() << ' ' << 
          rule.name_() << ' ' << 
          rule.retval_();
#ifdef PRINT_DBG
        std::cout << ":EGSE:";        
#endif
      }
      static void emitNested(crule_t const& rule, iterator_t begin, iterator_t end) {
#ifdef PRINT_DBG
        std::cout << ":ENSB:";
#endif
        std::ostringstream os;
        os << rule.type_() << ' ' << rule.name_() << '_' << rule.type_() << ' ' 
           << rule.retval_() << std::string(begin, end) << std::endl;
        os << rule.name_() << '_' << rule.type_() << ' ' << rule.name_();        
        rule.retval_() = os.str();
#ifdef PRINT_DBG
        std::cout << ":ENSE:";
#endif
      }

      static std::string::iterator nameBegin(crule_t const& rule) {
        return rule.name_().begin();
      }
      static std::string::iterator nameEnd(crule_t const& rule) {
        return rule.name_().end();
      }
    };
  };
}

#endif
