/********************************************************************************

Copyright: Acclaim Systems Ltd 2004
Author:    rskinner
 
No claims are made concerning the suitability of this source code for any purpose. 
All source is provided 'as is' without express or implied warranty of any kind.
Acclaim Systems Ltd disclaims any/all implied warranties of merchantability, 
non-infringement or fitness for any particular purpose. In no event will Acclaim
Systems Ltd be liable for any special, indirect, incidental, consequential or 
any other damages resulting from loss of use, data or profits arising out of or 
in connection with the use or performance of this source code.

License:   Free to use and abuse for non-profit activities so long as the 
           attribution and disclaimer above are not altered. For-profit use 
           is prohibited without express written permission.
 
*******************************************************************************/

#ifndef H2D_TRANSFORM_H
#define H2D_TRANSFORM_H

#include "h2d.h"
#include "h2d_d.h"

#include <boost/spirit/actor/assign_actor.hpp>
#include <boost/bind.hpp>

#include <vector>
#include <string> 

////////////////////////////////////////////////////////////
//
// Hodge-podge of various odds n sods that need to be messed 
// around to be D compatible.
//
// Starting with....
//
// 1. sizeof(xxx) -> xxx.sizeof
// 2. alias ... <dtype>; -> 
//    removal of aliases to D types.
// 3. removal of const as a type modifier - actually we'll 
//    just be removing const in function declarations to start 
//    with. Also removes in,out,inout from functions decls - 
//    these are sometimes used as parameter names in C. Parameter
//    names shouldn't be missed unless they are in inline
//    function definitions - in which case we are screwed.
//

namespace h2d {

  struct ParensClosure : public boost::spirit::closure<ParensClosure,
                         std::string> {

    member1 retval_;    
  };

  struct TransformGrammar : public boost::spirit::grammar<TransformGrammar> {

    template<typename T>
    struct definition {

      typedef boost::spirit::rule<T> rule_t;
      typedef boost::spirit::rule<T, ParensClosure::context_t> crule_t;
      typedef typename T::iterator_t iterator_t;

      definition(TransformGrammar const& self) {

        // sizeof(xxx) -> xxx.sizeof
        sizeof_ = (str_p("sizeof") >> *space_p >> 
                     confix_p('(' >> *space_p, 
                              ((+print_p)[boost::spirit::assign_a(sizeofType_)]),
                              *space_p >> ')'))[boost::bind(&definition<T>::sizeofEmit, this)];

        // Sometimes C headers define typedefs to D type names
        // these must be removed
        dtypedef_ = str_p("alias") >> +space_p >> ((+(anychar_p - dtypes_)) | dtypes_) >> +space_p >> dtypes_ >> ch_p(';');

        // Remove 'const' from function definitions - first define a rule to 
        // remove const from within parenthese
        boost::function<void(char_t)> pcFn = boost::bind(&definition<T>::captureC, boost::ref(parens_), _1);
        boost::function<void(std::string const&)> psFn = boost::bind(&definition<T>::captureS, boost::ref(parens_), _1);
        boost::function<void(iterator_t, iterator_t)> pbeFn = boost::bind(&definition<T>::captureBE, boost::ref(parens_), _1, _2);

        keywords_ = (str_p("const") | str_p("in") | str_p("out") | str_p("inout")) >> eps_p(~chset_p("a-zA-Z0-9_"));
        delimiters_ = chset_p("&*,") | space_p;
        parens_ = 
          ch_p('(')[pcFn] 
            >> 
          *(parens_[psFn] | (+delimiters_)[pbeFn] | keywords_ | (+(anychar_p - (chset_p("()") | (delimiters_ >> keywords_))))[pbeFn]) 
            >> 
          ch_p(')')[pcFn];
        // Now define a rule to use parens_ on function declarations/definitions
        functions_ = 
          // Used to check for dtype return specification but this isn't a valid assumption
          //          (dtypes_ >> +space_p >> +(alnum_p | ch_p('_')) >> *space_p)[rsFn] 
          // Nor can we assume it will be a simple function definition/declaration (think ptr to fn)
          //          (+(alnum_p | ch_p('_')) >> *space_p)[rsFn] 
          //          >> 
          
          // Optional 'const' for return value - dropped
          (!(str_p("const") >> (+space_p)[boost::bind(&definition<T>::captureBE, boost::ref(functions_), _1, _2)]))
          >>
          // followed by return value type and...
          ((+graph_p >> +(space_p | ch_p('*')))[boost::bind(&definition<T>::captureBE, boost::ref(functions_), _1, _2)] 
          >> 
           // either a function name 
           ((+(graph_p - ch_p('(')) >> *space_p)[boost::bind(&definition<T>::captureBE, boost::ref(functions_), _1, _2)] 
           | // or a function pointer definition
           parens_[boost::bind(&definition<T>::captureS, boost::ref(functions_), _1)]))
          >>
          parens_[boost::bind(&definition<T>::captureS, boost::ref(functions_), _1)]
          >> 
          (*space_p >> chset_p("{;"))[boost::bind(&definition<T>::echoS, boost::ref(functions_), _1, _2)];
        
        
        rule_ = sizeof_ | dtypedef_ | functions_;
      }
      
      rule_t const& start() const { return rule_; }
    private:
      void sizeofEmit() const {
        std::cout << sizeofType_ << ".sizeof";
      }
      static void captureC(crule_t const& rule, char_t c) {
        rule.retval_() += c;
      }
      static void captureS(crule_t const& rule, std::string const& str) {
        rule.retval_() += str;
      }
      static void captureBE(crule_t const& rule, iterator_t begin, iterator_t end) {
        rule.retval_() += std::string(begin, end);
      }
      static void echoS(crule_t const& rule, iterator_t begin, iterator_t end) {
        cout << rule.retval_() << std::string(begin, end);
      }

      rule_t        rule_;

      rule_t        sizeof_;
      rule_t        dtypedef_;
      crule_t       parens_;
      crule_t       functions_;
      rule_t        keywords_;
      rule_t        delimiters_;

      std::string   sizeofType_;
      DTypesGrammar dtypes_;
      std::string   parensContent_;
    };
    
  };
}

#endif
